﻿using System.Collections.ObjectModel;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.WpfDesktopEdition;

namespace SnapToLayer2
{
    public enum ToleranceCoordinates { World, Screen };
    class SnapToLayerTrackInteractiveOverlay : TrackInteractiveOverlay
    {
        private InMemoryFeatureLayer toSnapInMemoryFeatureLayer;
        private float tolerance;
        private DistanceUnit toleranceUnit;
        private GeographyUnit geographyUnit;

        private RectangleShape currentWorldExtent;
        private float mapWidth;
        private float mapHeight;
        private ToleranceCoordinates toleranceType;

        //InMemoryFeatureLayer for the layer to be snapped to.
        public InMemoryFeatureLayer ToSnapInMemoryFeatureLayer
        {
            get { return toSnapInMemoryFeatureLayer; }
            set { toSnapInMemoryFeatureLayer = value; }
        }

        public ToleranceCoordinates ToleranceType
        {
            get { return toleranceType; }
            set { toleranceType = value; }
        }

        public float Tolerance
        {
            get { return tolerance; }
            set { tolerance = value; }
        }

        public DistanceUnit ToleranceUnit
        {
            get { return toleranceUnit; }
            set { toleranceUnit = value; }
        }

        protected override void OnVertexAdding(VertexAddingTrackInteractiveOverlayEventArgs e)
        {
            if (currentWorldExtent != null)
            {
                PointShape currentVertex = new PointShape(e.AddingVertex);
                PointShape snapPointShape = new PointShape();
                if (toleranceType == ToleranceCoordinates.Screen)
                {
                    snapPointShape = FindNearestSnappingPointPixel(currentVertex);
                }
                else
                {
                    snapPointShape = FindNearestSnappingPoint(currentVertex);
                }
                if (snapPointShape != null)
                {
                    e.AddingVertex = new Vertex(snapPointShape.X, snapPointShape.Y);

                }
            }
            base.OnVertexAdding(e);
        }

        protected override void OnTrackEnded(TrackEndedTrackInteractiveOverlayEventArgs e)
        {
            // Update the last vertex
            Vertex lastVertex = this.Vertices[this.Vertices.Count - 2];
            PointShape lastPoint = new PointShape(lastVertex.X, lastVertex.Y);
            PointShape snapPointShape = new PointShape();
            if (toleranceType == ToleranceCoordinates.Screen)
            {
                snapPointShape = FindNearestSnappingPointPixel(lastPoint);
            }
            else
            {
                snapPointShape = FindNearestSnappingPoint(lastPoint);
            }

            if (snapPointShape != null)
            {
                PolygonShape trackPolygon = e.TrackShape as PolygonShape;
                // find the tracked shape and update it to snap
                Feature removedFeature = null;
                for (int i = 0; i < this.TrackShapeLayer.InternalFeatures.Count; i++)
                {
                    if (this.TrackShapeLayer.InternalFeatures[i].GetShape().Id == trackPolygon.Id)
                    {
                        removedFeature = this.TrackShapeLayer.InternalFeatures[i];
                    }
                }

                // Update trackShapeLayer
                if (removedFeature != null)
                {
                    this.TrackShapeLayer.InternalFeatures.Remove(removedFeature);
                    trackPolygon.OuterRing.Vertices[trackPolygon.OuterRing.Vertices.Count - 2] = new Vertex(snapPointShape.X, snapPointShape.Y);
                    this.TrackShapeLayer.InternalFeatures.Add(new Feature(trackPolygon));
                }
            }
            base.OnTrackEnded(e);
        }

        protected override InteractiveResult MouseMoveCore(InteractionArguments interactionArguments)
        {
            currentWorldExtent = interactionArguments.CurrentExtent;
            mapHeight = interactionArguments.MapHeight;
            mapWidth = interactionArguments.MapWidth;
            return base.MouseMoveCore(interactionArguments);
        }

        protected override void DrawTileCore(GeoCanvas geoCanvas)
        {
            //Sets the geography Unit used in FindNearestSnappingPoint function
            geographyUnit = geoCanvas.MapUnit;

            currentWorldExtent = geoCanvas.CurrentWorldExtent;
            mapWidth = geoCanvas.Width;
            mapHeight = geoCanvas.Height;

            //Draws the Edit Shapes as default.
            Collection<SimpleCandidate> labelsInAllLayers = new Collection<SimpleCandidate>();
            //EditShapesLayer.Open();
            TrackShapeLayer.Open();
            //EditShapesLayer.Draw(geoCanvas, labelsInAllLayers);
            TrackShapeLayer.Draw(geoCanvas, labelsInAllLayers);
            geoCanvas.Flush();
        }

        private PointShape FindNearestSnappingPointPixel(PointShape targetPointShape)
        {
            toSnapInMemoryFeatureLayer.Open();
            Collection<Feature> toSnapInMemoryFeatures = toSnapInMemoryFeatureLayer.FeatureSource.GetFeaturesNearestTo(targetPointShape, GeographyUnit.Meter, 1, ReturningColumnsType.AllColumns);
            toSnapInMemoryFeatureLayer.Close();

            if (toSnapInMemoryFeatures.Count == 1)
            {
                MultipolygonShape multipolygonShape = (MultipolygonShape)toSnapInMemoryFeatures[0].GetShape();

                foreach (PolygonShape polygonShape in multipolygonShape.Polygons)
                {
                    foreach (Vertex vertex in polygonShape.OuterRing.Vertices)
                    {
                        PointShape toSnapPointShape = new PointShape(vertex);
                        float screenDistance = ExtentHelper.GetScreenDistanceBetweenTwoWorldPoints(currentWorldExtent, toSnapPointShape, targetPointShape, mapWidth, mapHeight);

                        if (screenDistance <= tolerance)
                        {
                            return new PointShape(toSnapPointShape.X, toSnapPointShape.Y);
                        }
                    }
                }
            }
            return null;
        }

        //Function to find if dragged control point is within the tolerance of a vertex of layer in world coordinates.
        private PointShape FindNearestSnappingPoint(PointShape targetPointShape)
        {
            toSnapInMemoryFeatureLayer.Open();
            Collection<Feature> toSnapInMemoryFeatures = toSnapInMemoryFeatureLayer.FeatureSource.GetFeaturesNearestTo(targetPointShape, GeographyUnit.Meter, 1, ReturningColumnsType.AllColumns);
            toSnapInMemoryFeatureLayer.Close();

            if (toSnapInMemoryFeatures.Count == 1)
            {
                PolygonShape polygonShape = (PolygonShape)toSnapInMemoryFeatures[0].GetShape();

                foreach (Vertex vertex in polygonShape.OuterRing.Vertices)
                {
                    PointShape toSnapPointShape = new PointShape(vertex);
                    double Distance = toSnapPointShape.GetDistanceTo(targetPointShape, geographyUnit, toleranceUnit);

                    if (Distance <= tolerance)
                    {
                        return new PointShape(toSnapPointShape.X, toSnapPointShape.Y);
                    }
                }
            }
            return null;
        }
    }
}
